"""
WeatherAPI types.
"""

import json
import pickle
from datetime import datetime as dt

from . import raw

__all__ = [
    'Location',
    'HourlyPoint',
    'DayPoint',
    'AstroPoint',
    'DailyPoint',
    'IpPoint',
    'SportsPoint'
]

class BasePoint:
    """Abstract class representing a single data point."""
    @classmethod
    def from_raw(cls, data):
        """Return new instance of child from json converted `weatherapi` response."""
        # Unpacking uses the dict's keys are keyword arguments
        return cls(**data)

    @classmethod
    def from_json(cls, data):
        """Return new instance of child from raw `weatherapi` response."""
        return cls(**json.loads(data))

    def to_bytes(self):
        """Return a bytes-like object."""
        return pickle.dumps(self)

class BasePointLoc:
    """``BasePoint`` with extra location parameter."""
    @classmethod
    def from_raw(cls, location, data):
        """Return new instance of child from json converted `weatherapi` response."""
        return cls(location, **data) # Unpacking uses the dict's keys are keyword arguments

    @classmethod
    def from_json(cls, location, data):
        """Return new instance of child from raw `weatherapi` response."""
        return cls(location, **json.loads(data))

    def to_bytes(self):
        """Return a bytes-like object."""
        return pickle.dumps(self)

# Inheritance only to implement `from_raw` and `from_json` automatically for all subclasses

class Location(BasePoint):
    """
    Represents location data such as coordinates, time zone, region, at a particular time.

    :var lat: :class:`float`
    :var lon: :class:`float`
    :var name: :class:`str`
    :var region: :class:`str`
    :var tz_id: :class:`str`
    :var localtime: :class:`datetime.datetime`
    """
    def __init__(self,
        lat, lon, name, region=None, country=None, tz_id=None, localtime=None,
        **kwargs
    ):
        self.lat = lat
        self.lon = lon
        self.name = name
        self.region = region
        self.country = country
        self.tz_id = tz_id
        self.localtime = dt.strptime(localtime, "%Y-%m-%d %H:%M") if localtime else None

class HourlyPoint(BasePointLoc):
    """
    Represents weather data at a particular time in some location.

    :var location: :class:`Location`
    :var time: :class:`datetime.datetime`
    :var temp_c: :class:`types.Cel`
    :var feelslike_c: :class:`types.Cel`
    :var windchill_c: :class:`types.Cel`
    :var heatindex_c: :class:`types.Cel`
    :var dewpoint_c: :class:`types.Cel`
    :var condition: :class:`dict`
    :var wind_kph: :class:`types.Km`
    :var gust_kph: :class:`types.Km`
    :var wind_degree: :class:`int`
    :var wind_dir: :class:`str`
    :var pressure_mb: :class:`types.Mb`
    :var precip_mm: :class:`types.Mm`
    :var will_it_rain: :class:`int`
    :var will_it_snow: :class:`int`
    :var chance_of_rain: :class:`str`
    :var chance_of_snow: :class:`str`
    :var humidity: :class:`int`
    :var cloud: :class:`int`
    :var is_day: :class:`bool`
    :var uv: :class:`float`
    :var vis_km: :class:`types.Km`
    """
    def __init__(
        self, location,
        temp_c, feelslike_c,
        condition,
        wind_kph, wind_degree, wind_dir, gust_kph,
        pressure_mb, precip_mm, humidity,
        cloud, is_day, uv,
        last_updated=None, time=None,
        windchill_c=None, heatindex_c=None, dewpoint_c=None,
        will_it_rain=None, will_it_snow=None,
        chance_of_rain=None, chance_of_snow=None,
        vis_km=None,
        **kwargs
    ):
        if isinstance(location, Location):
            self.location = location
        else:
            self.location = Location.from_raw(location)

        self.time = dt.strptime(last_updated if not time else time, "%Y-%m-%d %H:%M") \
                    if (last_updated or time) else None

        self.temp_c = raw.Cel(temp_c)
        self.feelslike_c = raw.Cel(feelslike_c)

        self.windchill_c = raw.Cel(windchill_c)
        self.heatindex_c = raw.Cel(heatindex_c)
        self.dewpoint_c = raw.Cel(dewpoint_c)

        self.condition = condition

        self.wind_kph = raw.Km(wind_kph)
        self.gust_kph = raw.Km(gust_kph)

        self.wind_degree = wind_degree
        self.wind_dir = wind_dir

        self.pressure_mb = raw.Mb(pressure_mb)
        self.precip_mm = raw.Mm(precip_mm)

        self.will_it_rain = will_it_rain
        self.will_it_snow = will_it_snow
        self.chance_of_rain = chance_of_rain
        self.chance_of_snow = chance_of_snow

        self.humidity = humidity
        self.cloud = cloud

        self.is_day = bool(is_day)

        self.uv = uv
        self.vis_km = raw.Km(vis_km)

class DayPoint(BasePointLoc):
    """
    The total conditions per day.

    :var location: :class:`Location`
    :var maxtemp_c: :class:`types.Cel`
    :var mintemp_c: :class:`types.Cel`
    :var avgtemp_c: :class:`types.Cel`
    :var condition: :class:`dict`
    :var maxwind_kph: :class:`types.Km`
    :var totalprecip_mm: :class:`types.Mm`
    :var avgvis_km: :class:`types.Km`
    :var avghumidity: :class:`int`
    :var uv: :class:`float`
    """
    def __init__(
        self, location,
        maxtemp_c, mintemp_c, avgtemp_c, maxwind_kph,
        totalprecip_mm, avgvis_km, avghumidity, condition, uv,
        **kwargs
    ):
        if isinstance(location, Location):
            self.location = location
        else:
            self.location = Location.from_raw(location)

        self.maxtemp_c = raw.Cel(maxtemp_c)
        self.mintemp_c = raw.Cel(mintemp_c)
        self.avgtemp_c = raw.Cel(avgtemp_c)

        self.condition = condition

        self.maxwind_kph = raw.Km(maxwind_kph)

        self.totalprecip_mm = raw.Mm(totalprecip_mm)
        self.avgvis_km = raw.Km(avgvis_km)
        self.avghumidity = avghumidity

        self.uv = uv

class AstroPoint(BasePointLoc):
    """
    Astronomy information.

    :var location: :class:`Location`
    :var sunrise: :class:`str`
    :var sunset: :class:`str`
    :var moonrise: :class:`str`
    :var moonset: :class:`str`
    :var moon_phase: :class:`str`
    """
    def __init__(self, location, sunrise, sunset, moonrise, moonset, moon_phase, **kwargs):
        if isinstance(location, Location):
            self.location = location
        else:
            self.location = Location.from_raw(location)

        self.sunrise = sunrise
        self.sunset = sunset

        self.moonrise = moonrise
        self.moonset = moonset

        self.moon_phase = moon_phase

class DailyPoint(BasePointLoc):
    """
    All information per day, inlcuding hourly info.

    :var location: :class:`Location`
    :var day: :class:`DayPoint`
    :var astro: :class:`AstroPoint`
    :var hour: ``list[HourlyPoint]``
    """
    def __init__(self, location, day, astro, hour):
        self.location = location if isinstance(location, Location) else Location.from_raw(location)

        self.day = day if isinstance(day, DayPoint) else DayPoint.from_raw(location, day)

        self.astro = astro

        self.hour = [
            i if isinstance(i, HourlyPoint) else HourlyPoint.from_raw(location, i)
            for i in hour
        ]

class IpPoint(BasePoint):
    """
    IP Address information.

    :var location: :class:`Location`
    :var ip: :class:`str`
    :var type: :class:`str`
    """
    def __init__(self,
        ip, type,
        country_name,
        city, region,
        lat, lon, tz_id,
        **kwargs
    ):
        self.location = Location(lat, lon, city, region=region, country=country_name, tz_id=tz_id)
        self.ip = ip
        self.type = type

class SportsPoint(BasePoint):
    """
    Information about a sports event, such as stadium, region, start time, etc.

    :var stadium: :class:`str`
    :var country: :class:`str`
    :var region: :class:`str`
    :var tournament: :class:`str`
    :var start: :class:`datetime.datetime`
    :var match: :class:`str`
    """
    def __init__(self, stadium, country, region, tournament, start, match):
        self.stadium = stadium
        self.country = country # / Might wrap in a `Location` object
        self.region = region   # /
        self.tournament = tournament
        self.start = dt.strptime(start, "%Y-%m-%d %H:%M") if start else None
        self.match = match



"""
Cache utility.
The primary purpose of this is to reduce the number of required API calls.
"""

import os
import sys

import zlib
import pickle

from pathlib import Path

__all__ = [
    'CacheManager',
    'BufferedCacheManager',
    'FileCacheManager'
]

class CacheManager:
    """
    Abstract class representing a cache manager.
    """

    def __init__(self, path):
        pass

    def find_all(self):
        """Find all stored cache."""
        return None

    def read(self, name):
        """Reads cache with ``name`` if it exists."""
        return None

    def dump(self, name, data):
        """Save data into cache."""
        return None

    def cleanup(self, name):
        """Cleanup all cache with a ``name``"""
        return None

    def debug_size(self):
        """Get size of cache for debugging."""
        return sys.getsizeof(self)


class BufferedCacheManager(CacheManager):
    """
    Buffered Cache Manager implementation. Keeps track of cache in memory.
    """

    def __init__(self, path=None):
        self._buf = {}

        super().__init__(path)

    def find_all(self):
        """
        Find all cache stored in memory.

        :rtype: :class:`generator`
        """
        return (
            i
            for i in self._buf
        )

    def read(self, name):
        """Reads cache with ``name`` if it exists."""
        if name in self._buf:
            return pickle.loads(zlib.decompress(self._buf[name]))

        return None

    def dump(self, name, data):
        """Save data into cache."""
        self._buf[name] = zlib.compress(pickle.dumps(data))

    def cleanup(self, name):
        """Cleanup all cache with a ``name``. Supports wildcard (*) deletion."""

        els = name.split('*') # splits across *

        need_remove = []

        for i in self._buf:
            for n, j in enumerate(els):
                i = i.rstrip('.dat')
                if not (
                    j == '' or (j in i and
                        (i.index(j) >= i.index(els[0 if n < 1 else n - 1]))
                        )
                    ):
                    break

                    # Explanation for above check
                    # ---------------------------
                    # `els` is a list of all components split across *.
                    # We check if each component of `else` is in
                    # the file name being checked (`j in i`).
                    # If the component is empty, we can skip directly.
                    # If it is, we make sure its after the previous
                    # component (second check).
            else:
                need_remove.append(i)

        for i in need_remove:
            del self._buf[i]

    def debug_size(self):
        """
        Get size of internal cache for debugging.

        :returns: Cache size in memory in bytes.
        """
        return sys.getsizeof(self._buf)

class FileCacheManager(CacheManager):
    """
    File based Cache Manager implementation. Keeps track of and gets/updates data from cache files.

    :param path: Path to the "cache directory". Cache files will be stored here.
    """

    def __init__(self, path):
        self._path = path

        super().__init__(path)

        Path(path).mkdir(parents=True, exist_ok=True)  # Creates cache folder

        ## Cache is identified with its `name` attribute. Cache can be read by keeping track of this
        ## value and reading it with `read` later on.

    @property
    def path(self):
        """
        Returns the directory to which cache files are being stored.

        :rtype: :class:`str`
        """
        return self._path

    def find_all(self):
        """
        Return a list of all tracked cache files.

        :rtype: :class:`generator`
        """
        return (
            i.rstrip('.dat')
            for i in os.listdir(self._path)
        )

    def read(self, name):
        """Tries to read cache with ``name``. Returns ``None`` if no such file is found."""

        try:
            # Cache is stored as an object in a binary file,
            # which can be loaded into an object directly later on.
            with open(f"{self._path}/{name}.dat", "rb") as f:
                return pickle.loads(zlib.decompress(pickle.load(f)))

        except pickle.PickleError:
            pass
        except FileNotFoundError:
            pass

        return None

    def dump(self, name, data):
        """Writes data to a cache file with ``name``. ``name`` must be kept track of manually."""
        with open(f"{self._path}/{name}.dat", "wb") as f:
            pickle.dump(zlib.compress(pickle.dumps(data)), f)

    def cleanup(self, name):
        """Cleans up all cache files with a given ``name``. Supports wildcard (*) deletion."""

        els = name.split('*') # splits across *

        # Removes all cache files matching the `name` pattern. `*` represents any set of characters.

        try:
            for i in os.listdir(self._path):
                for n, j in enumerate(els):
                    k = i.rstrip('.dat')
                    if not (
                        j == '' or (j in i and
                            (k.index(j) >= k.index(els[0 if n < 1 else n - 1]))
                            )
                        ):
                        break

                        # Explanation for above check
                        # ---------------------------
                        # `els` is a list of all components split across *.
                        # We check if each component of `else` is in
                        # the file name being checked (`j in i`).
                        # If the component is empty, we can skip directly.
                        # If it is, we make sure its after the previous
                        # component (second check).
                else:
                    os.remove(f"{self._path}/{i}") # Delete the actual file

        except FileNotFoundError:
            pass



"""
All exceptions that can be raised during handling of a weatherapi request.
"""

# Functions that make a request to weatherapi.com have the possibility
# of raising one of these errors depending on the response.

# We use these custom error classes to make the error code more readable
# and easier to handle

__all__ = [
    'WeatherApiError',
    'NoApiKey',
    'InvalidApiKey',
    'QuotaExceeded',
    'ApiKeyDisabled',
    'QueryNotProvided',
    'InvalidRequestUrl',
    'InvalidLocation',
    'InternalError'
]

class WeatherApiError(Exception):
    """
    Raised when an unknown weatherapi error is encountered.

    :var internal_code: :class:`int`
        The error code returned by weatherapi.
    :var message: :class:`str`
        A more detailed description of the error.
    """

    def __init__(self, message, internal_code):
        self.internal_code = internal_code
        self.message = message
        super().__init__(message)

class NoApiKey(WeatherApiError):
    """Raised when no weatherapi key has been provided in a request."""

class InvalidApiKey(WeatherApiError):
    """Raised when an invalid weatherapi key has been used for a request."""

class QuotaExceeded(WeatherApiError):
    """Raised when monthly requests limit has been reached."""

class ApiKeyDisabled(WeatherApiError):
    """Raised when weatherapi key used for a request has been disabled."""

class QueryNotProvided(WeatherApiError):
    """Raised when a query parameter has not been provided for a request."""

class InvalidRequestUrl(WeatherApiError):
    """Raised when weatherapi request url is invalid."""

class InvalidLocation(WeatherApiError):
    """Raised when location request is not found."""

class InternalError(WeatherApiError):
    """Raised when an internal weatherapi error is encountered."""



"""
Utility classes for easy conversions.
"""

from dataclasses import dataclass

# The use of dataclasses here
# ---------------------------
# The dataclasses module provides a decorator and
# functions for automatically generating special methods
# such as __repr__() for user-defined classes
# [docs](https://docs.python.org/3/library/dataclasses.html).

__all__ = [
    'Km',
    'Mm',
    'Mb',
    'Cel'
]

@dataclass(frozen=True)
class Km:
    """Kilometer."""
    val: float

    def miles(self):
        """Equivalent miles."""
        return self.val * 0.6213

    def __repr__(self):
        return f"{self.val}"

@dataclass(frozen=True)
class Mm:
    """Milimeters."""
    val: float

    def inches(self):
        """Equivalent inches."""
        return self.val * 0.0393

    def __repr__(self):
        return f"{self.val}"

@dataclass(frozen=True)
class Mb:
    """Milibar."""
    val: float

    def inches(self):
        """Equivalent inches of Hg."""
        return self.val * 0.02952

    def __repr__(self):
        return f"{self.val}"

@dataclass(frozen=True)
class Cel:
    """Celsius."""
    val: float

    def fahrenheit(self):
        """Equivalent farenheit."""
        return (self.val * 1.8) + 32

    def kelvin(self):
        """Equivalent farenheit."""
        return self.val + 273.15

    def __repr__(self):
        return f"{self.val}"



"""
Setup file. `pip` uses this to install `speck`.
"""

import os
import re

from setuptools import setup

# Utils ----------------

def readf(fname):
    print(os.path.join(os.path.dirname(__file__), fname))
    with open(os.path.join(os.path.dirname(__file__), fname), 'r') as f:
        return f.read()

# From discord.py
# https://github.com/Rapptz/discord.py/blob/master/setup.py#L10

def get_version(fname):
    return re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', readf(fname), re.MULTILINE).group(1)
    # +------------------------------------------------------------+
    # | Regex                                                      |
    # +------------------------------------------------------------+
    # | - re.MULTILINE tells `search` to compare each line         |
    # |     instead of the whole text.                             |
    # | - \s - whitespace character                                |
    # | - [\'"] - either ' or "                                    |
    # | - ([^\'"]*) - characters that are NOT ' or "               |
    # |     .group(1) will return the first subgroup of the match  |
    # |     subgroups are enclosed in `()`                         |
    # +------------------------------------------------------------+

# Requirements ---------

version = get_version('speck/__init__.py')

install_requires = list(readf('requirements.txt').splitlines())
extras_require = {
    'docs': [
        'sphinx',
        'sphinx-press-theme'
    ]
}

packages = [
    "speck",
    "speck.types",
    "speck.etc"
]
package_data = {
    '': ['LICENSE', 'README.md', 'etc/*']
}

# ----------------------

setup(
    name = "speck-wrapper",
    version = version,
    license = "MIT",
    author = "Nevin Jose, Sachin Cherian",
    description = "A simple wrapper and frontend for weatherAPI.com",
    long_description = readf('README.md'),
    long_description_content_type = 'text/markdown',
    url = "https://github.com/schctl/speck",
    packages = packages,
    package_data = package_data,
    include_package_data = True,
    install_requires = install_requires,
    extras_require = extras_require,
    python_requires = ">=3.6.0"
)



"""
WeatherAPI client implementation.
Use this to make requests to weatherapi.com.
"""

import os
import json

from datetime import datetime as dt

import requests

from . import cache
from . import errors
from . import types

__all__ = [
    'Client'
]

class Client:
    """
    Represents a connection to weatherapi.com.
    Use this class to interact with the weatherapi API.

    :var session: A `requests.Session` object. Requests are made with this (``session.get``).
    """

    BASE = "https://api.weatherapi.com/v1"

    def __init__(self, token, use_cache=False, cache_file=False, cache_path='.cache'):
        self._token = token
        self.session = requests.Session()

        if use_cache:
            if cache_file:
                self.cache = cache.FileCacheManager(cache_path)
            else:
                self.cache = cache.BufferedCacheManager(cache_path)
        else:
            self.cache = cache.CacheManager(cache_path)

        # This looks for the cities list file
        with open(
            os.path.join(os.path.dirname(__file__), 'etc/cities_p.json'),
            'r', encoding='utf-8'
        ) as f:
            self.cities = json.loads(f.read())

    @staticmethod
    def __is_error_code(response):
        """
        Convert weatherapi.com provided error code to Error Type.

        :param response: The raw weatherapi.com response.
        """

        if "error" in response:
            code = response['error']['code']
            message = response['error']['message']

            # These are hardcoded
            if code == 1002:
                return errors.NoApiKey(message, code)
            elif code == 1003:
                return errors.QueryNotProvided(message, code)
            elif code == 1005:
                return errors.InvalidRequestUrl(message, code)
            elif code == 1006:
                return errors.InvalidLocation(message, code)
            elif code == 1008:
                return errors.ApiKeyDisabled(message, code)
            elif code == 2006:
                return errors.InvalidApiKey(message, code)
            elif code == 2007:
                return errors.QuotaExceeded(message, code)
            elif code == 2008:
                return errors.ApiKeyDisabled(message, code)
            elif code == 9999:
                return errors.InternalError(message, code)
            else:
                return errors.WeatherApiError(message, code)

        return

    def __make_request(self, endpoint, parameters):
        """Private method to make a request to ``weatherapi.com``."""

        try:
            # Does the acutal request
            return self.session.get(f"{self.BASE}/{endpoint}{parameters}",
                                    timeout=(4, 4)).json()
        except Exception as e:
            raise errors.InternalError(f"Unable to fetch data at this time: {e}", 9999)

    def find_city(self, loc):
        """
        Try to find a city with a match from a known list of locations.
        This method is highly unlikely to be used frequently, since weatherAPI tries
        to interpret location names even if incorrect, but is provided anyway.
        """

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        return [  # Generates a list of city names containing the string `loc`
            i for i in self.cities if loc.lower() in i['name'].lower()
        ]

    def current(self, loc):
        """
        Get current weather conditions in a location.

        :param loc: Query location to find data for. It could be following:\n
             - **Latitude and Longitude (Decimal degree)**. *e.g:'48.8567,2.3508'*,\n
             - **city name** *e.g: 'Paris'*,\n
             - **US zip** *e.g: '10001'*,\n
             - **UK postcode** *e.g: 'SW1'*,\n
             - **Canada postal code** *e.g: 'G2J'*,\n
             - **metar:<metar code>** *e.g: 'metar:EGLL'*,\n
             - **iata:<3 digit airport code>** *e.g: 'iata:DXB'*,\n
             - **auto:ip IP lookup** *e.g: 'auto:ip'*,\n
             - **IP address (IPv4 and IPv6 supported)** *e.g: '100.0.0.1*'

        :rtype: :class:`types.HourlyPoint`
        """

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        mode = f"current-{loc.lower()}-now-{dt.now().strftime('%Y-%m-%d-%H-%M')[:-1]}"

        # https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior

        # If cache exists (not None), it will be read and an `HourlyPoint` object will be returned

        n = self.cache.read(mode)
        if n:
            return n

        response = self.__make_request('current.json', f'?key={self._token}&q={loc}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        # Creates the `HourlyPoint` object
        data = types.HourlyPoint.from_raw(response["location"], response["current"])

        self.cache.cleanup(mode.split('-now-')[0] + '-now-*') # Discard any old cache
        self.cache.dump(mode, data) # Writes cache

        return data

        # The same pattern is followed for all other API methods implemented.

    def forecast(self, loc, days=3):
        """
        Get weather forecast for a location.

        :param loc: See docs on method ``current``.
        :param days: Number of days to restrict the forecast for.
            WeatherAPI allows up to 10 days, but it in practice the maximum is 3.

        :returns: A tuple with the current weather, and a list of forecasted days.
        :rtype: ``(types.HourlyPoint, list[types.DailyPoint])``
        """

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        mode = f"forecast-{loc.lower()}-now-{dt.now().strftime('%Y-%m-%d-%H-%M')[:-1]}"

        n = self.cache.read(mode)
        if n:
            return n

        response = self.__make_request('forecast.json', f'?key={self._token}&q={loc}&days={min(days, 10)}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        # `data` here is a tuple of the current weather (`HourlyPoint`)
        # and a list of forecasted days (`DailyPoint`s).
        data = (
            types.HourlyPoint.from_raw(response["location"], response["current"]),
            [
                types.DailyPoint(response["location"], i["day"], i["astro"], i["hour"])
                for i in response["forecast"]["forecastday"]
            ]
        )

        self.cache.cleanup(mode.split('-now-')[0] + '-now-*')
        self.cache.dump(mode, data)

        return data

    def astronomy(self, loc):
        """
        Get astronomy information for a location.

        :param loc: See docs on method ``current``.

        :rtype: :class:`types.AstroPoint`
        """

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        mode = f"astro-{loc.lower()}-now-{dt.now().strftime('%Y-%m-%d')}"

        n = self.cache.read(mode)
        if n:
            return n

        response = self.__make_request('astronomy.json', f'?key={self._token}&q={loc}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        data = types.AstroPoint.from_raw(response["location"], response["astronomy"]["astro"])

        self.cache.cleanup(mode.split('-now-')[0] + '-now-*')
        self.cache.dump(mode, data)

        return data

    def ip_lookup(self, ip):
        """
        Get information for an IP address.

        :param ip: IPv6 or IPv4 string.

        :rtype: :class:`IpPoint`
        """

        if ip == '':
            raise errors.QueryNotProvided('IP cannot be empty.', 0)

        mode = f"iplookup-{ip}"

        n = self.cache.read(mode)
        if n:
            return n

        response = self.__make_request('ip.json', f'?key={self._token}&q={ip}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        data = types.IpPoint.from_raw(response)

        # self.cache.cleanup(mode) # We don't need this here
        self.cache.dump(mode, data)

        return data

    def search(self, loc):
        """
        Get a list of location objects based on query parameter.

        :param loc: See docs on method ``current``.

        :rtype: ``list[types.Location]``
        """

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        mode = f"search-{loc.lower()}"

        n = self.cache.read(mode)

        if n:
            return n

        response = self.__make_request('search.json', f'?key={self._token}&q={loc}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        data = [
            types.Location.from_raw(i)
            for i in response
        ]

        # self.cache.cleanup(mode) # We don't need this here
        self.cache.dump(mode, data)

        return data

    def timezone_info(self, loc):
        """
        Get timezone and associated information for a location.

        :param loc: See docs on method ``current``.

        :rtype: :class:`types.Location`
        """

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        # No cache

        response = self.__make_request('timezone.json', f'?key={self._token}&q={loc}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        return types.Location.from_raw(response["location"])

    def sports_lookup(self, loc):
        """
        Get listing of all upcoming sports events for football,
        cricket and golf. From the behaviour of the WeatherAPI Sports API,
        parameter `loc` doesn't actually matter but is required anyway.

        :param loc: See docs on method ``current``.

        :rtype: ``dict{str: list[SportsPoint]}``
        """
        mode = f"sports-{loc.lower()}-now-{dt.now().strftime('%Y-%m-%d')}"

        n = self.cache.read(mode)
        if n:
            return n

        response = self.__make_request('sports.json', f'?key={self._token}&q={loc}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        # `SportsPoint` contains data per sports event. It's not specific
        # to any type of sport, nor is it a collection. We use a dictionary
        # with the type of sport as a key, and list of `SportsPoints` as its value.
        data = {
            j: [types.SportsPoint.from_raw(i) for i in response[j]]
            for j in ['football', 'cricket', 'golf']
        }


        self.cache.cleanup(mode.split('-now-')[0] + '-now-*')
        self.cache.dump(mode, data)

        return data

    def history(self, loc, days):
        """
        Get weather history for a location. This hasn't been tested
        since it requires a paid weatherapi.com plan, but should work.

        :param loc: See docs on method ``current``.
        :param dt: Datetime string in the format `YYY-MM-DD`.
            Data starting from this date will be returned.

        :returns: A tuple with the current weather, and a list of historical
                  weather per day.
        :rtype: ``(types.HourlyPoint, list[types.DailyPoint])``
        """
        # weatherapi returns history data as an `Forecast` object as per their API.
        # https://www.weatherapi.com/docs/#apis-history

        if loc == '':
            raise errors.QueryNotProvided('Location cannot be empty.', 0)

        mode = f"history-{loc.lower()}-now-{dt.now().strftime('%Y-%m-%d-%H-%M')[:-1]}"

        n = self.cache.read(mode)
        if n:
            return n

        response = self.__make_request('history.json', f'?key={self._token}&q={loc}&days={min(days, 10)}')

        e = Client.__is_error_code(response)
        if e:
            raise e

        # `data` here is a tuple of the current weather (`HourlyPoint`)
        # and a list of historical days (`DailyPoint`s).
        data = (
            types.HourlyPoint.from_raw(response["location"], response["current"]),
            [
                types.DailyPoint(response["location"], i["day"], i["astro"], i["hour"])
                for i in response["forecast"]["forecastday"]
            ]
        )

        self.cache.cleanup(mode.split('-now-')[0] + '-now-*')
        self.cache.dump(mode, data)

        return data

    # Aliases ---------------------------------------

    def astro(self, *args, **kwargs):
        """Alias for ``astronomy``."""
        return self.astronomy(*args, **kwargs)

    def ip(self, *args, **kwargs):
        """Alias for ``ip_lookup``."""
        return self.ip_lookup(*args, **kwargs)

    def tz(self, *args, **kwargs):
        """Alias for ``timezone_info``."""
        return self.timezone_info(*args, **kwargs)

    def sports(self, *args, **kwargs):
        """Alias for ``sports_lookup``."""
        return self.sports_lookup(*args, **kwargs)
